/* -*- C++ -*-
 *
 * This file is a part of LEMON, a generic C++ optimization library
 *
 * Copyright (C) 2003-2006
 * Egervary Jeno Kombinatorikus Optimalizalasi Kutatocsoport
 * (Egervary Research Group on Combinatorial Optimization, EGRES).
 *
 * Permission to use, modify and distribute this software is granted
 * provided that this copyright notice appears in all copies. For
 * precise terms see the accompanying LICENSE file.
 *
 * This software is provided "AS IS" with no warranty of any kind,
 * express or implied, and with no claim as to its suitability for any
 * purpose.
 *
 */

#ifndef LEMON_GRAPH_EXTENDER_H
#define LEMON_GRAPH_EXTENDER_H

#include <lemon/invalid.h>
#include <lemon/error.h>

namespace lemon {

  template <typename _Base>
  class GraphExtender : public _Base {
  public:

    typedef _Base Parent;
    typedef GraphExtender Graph;

    typedef typename Parent::Node Node;
    typedef typename Parent::Edge Edge;

    int maxId(Node) const {
      return Parent::maxNodeId();
    }

    int maxId(Edge) const {
      return Parent::maxEdgeId();
    }

    Node fromId(int id, Node) const {
      return Parent::nodeFromId(id);
    }

    Edge fromId(int id, Edge) const {
      return Parent::edgeFromId(id);
    }

    Node oppositeNode(const Node &n, const Edge &e) const {
      if (n == Parent::source(e))
	return Parent::target(e);
      else if(n==Parent::target(e))
	return Parent::source(e);
      else
	return INVALID;
    }

  };

  template <typename _Base>
  class UGraphExtender : public _Base {
    typedef _Base Parent;
    typedef UGraphExtender Graph;

  public:

    typedef typename Parent::Edge UEdge;
    typedef typename Parent::Node Node;

    class Edge : public UEdge {
      friend class UGraphExtender;

    protected:
      // FIXME: Marci use opposite logic in his graph adaptors. It would
      // be reasonable to syncronize...
      bool forward;

      Edge(const UEdge &ue, bool _forward) :
        UEdge(ue), forward(_forward) {}

    public:
      Edge() {}

      /// Invalid edge constructor
      Edge(Invalid i) : UEdge(i), forward(true) {}

      bool operator==(const Edge &that) const {
	return forward==that.forward && UEdge(*this)==UEdge(that);
      }
      bool operator!=(const Edge &that) const {
	return forward!=that.forward || UEdge(*this)!=UEdge(that);
      }
      bool operator<(const Edge &that) const {
	return forward<that.forward ||
	  (!(that.forward<forward) && UEdge(*this)<UEdge(that));
      }
    };


    /// \brief Edge of opposite direction.
    ///
    /// Returns the Edge of opposite direction.
    Edge oppositeEdge(const Edge &e) const {
      return Edge(e,!e.forward);
    }

  public:
    /// \todo Shouldn't the "source" of an undirected edge be called "aNode"
    /// or something???
    using Parent::source;

    /// Source of the given Edge.
    Node source(const Edge &e) const {
      return e.forward ? Parent::source(e) : Parent::target(e);
    }

    /// \todo Shouldn't the "target" of an undirected edge be called "bNode"
    /// or something???
    using Parent::target;

    /// Target of the given Edge.
    Node target(const Edge &e) const {
      return e.forward ? Parent::target(e) : Parent::source(e);
    }

    Node oppositeNode(const Node &n, const UEdge &e) const {
      if( n == Parent::source(e))
	return Parent::target(e);
      else if( n == Parent::target(e))
	return Parent::source(e);
      else
	return INVALID;
    }

    /// \brief Directed edge from an undirected edge and a source node.
    ///
    /// Returns a (directed) Edge corresponding to the specified UEdge
    /// and source Node.
    ///
    Edge direct(const UEdge &ue, const Node &s) const {
      return Edge(ue, s == source(ue));
    }

    /// \brief Directed edge from an undirected edge.
    ///
    /// Returns a directed edge corresponding to the specified UEdge.
    /// If the given bool is true the given undirected edge and the
    /// returned edge have the same source node.
    Edge direct(const UEdge &ue, bool d) const {
      return Edge(ue, d);
    }

    /// Returns whether the given directed edge is same orientation as the
    /// corresponding undirected edge.
    ///
    /// \todo reference to the corresponding point of the undirected graph
    /// concept. "What does the direction of an undirected edge mean?"
    bool direction(const Edge &e) const { return e.forward; }


    using Parent::first;
    void first(Edge &e) const {
      Parent::first(e);
      e.forward=true;
    }

    using Parent::next;
    void next(Edge &e) const {
      if( e.forward ) {
	e.forward = false;
      }
      else {
	Parent::next(e);
	e.forward = true;
      }
    }

  public:

    void firstOut(Edge &e, const Node &n) const {
      Parent::firstIn(e,n);
      if( UEdge(e) != INVALID ) {
	e.forward = false;
      }
      else {
	Parent::firstOut(e,n);
	e.forward = true;
      }
    }
    void nextOut(Edge &e) const {
      if( ! e.forward ) {
	Node n = Parent::target(e);
	Parent::nextIn(e);
	if( UEdge(e) == INVALID ) {
	  Parent::firstOut(e, n);
	  e.forward = true;
	}
      }
      else {
	Parent::nextOut(e);
      }
    }

    void firstIn(Edge &e, const Node &n) const {
      Parent::firstOut(e,n);
      if( UEdge(e) != INVALID ) {
	e.forward = false;
      }
      else {
	Parent::firstIn(e,n);
	e.forward = true;
      }
    }
    void nextIn(Edge &e) const {
      if( ! e.forward ) {
	Node n = Parent::source(e);
	Parent::nextOut(e);
	if( UEdge(e) == INVALID ) {
	  Parent::firstIn(e, n);
	  e.forward = true;
	}
      }
      else {
	Parent::nextIn(e);
      }
    }

    void firstInc(UEdge &e, const Node &n) const {
      Parent::firstOut(e, n);
      if (e != INVALID) return;
      Parent::firstIn(e, n);
    }
    void nextInc(UEdge &e, const Node &n) const {
      if (Parent::source(e) == n) {
	Parent::nextOut(e);
	if (e != INVALID) return;
	Parent::firstIn(e, n);
      } else {
	Parent::nextIn(e);
      }
    }

    void firstInc(UEdge &e, bool &d, const Node &n) const {
      d = true;
      Parent::firstOut(e, n);
      if (e != INVALID) return;
      d = false;
      Parent::firstIn(e, n);
    }
    void nextInc(UEdge &e, bool &d) const {
      if (d) {
	Node s = Parent::source(e);
	Parent::nextOut(e);
	if (e != INVALID) return;
	d = false;
	Parent::firstIn(e, s);
      } else {
	Parent::nextIn(e);
      }
    }

    // Miscellaneous stuff:

    /// \todo these methods (id, maxEdgeId) should be moved into separate
    /// Extender

    // using Parent::id;
    // Using "using" is not a good idea, cause it could be that there is
    // no "id" in Parent...

    int id(const Node &n) const {
      return Parent::id(n);
    }

    int id(const UEdge &e) const {
      return Parent::id(e);
    }

    int id(const Edge &e) const {
      return 2 * Parent::id(e) + int(e.forward);
    }

    int maxNodeId() const {
      return Parent::maxNodeId();
    }

    int maxEdgeId() const {
      return 2 * Parent::maxEdgeId() + 1;
    }

    int maxUEdgeId() const {
      return Parent::maxEdgeId();
    }

    int maxId(Node) const {
      return maxNodeId();
    }

    int maxId(Edge) const {
      return maxEdgeId();
    }
    int maxId(UEdge) const {
      return maxUEdgeId();
    }

    int edgeNum() const {
      return 2 * Parent::edgeNum();
    }

    int uEdgeNum() const {
      return Parent::edgeNum();
    }

    Node nodeFromId(int id) const {
      return Parent::nodeFromId(id);
    }

    Edge edgeFromId(int id) const {
      return direct(Parent::edgeFromId(id >> 1), bool(id & 1));
    }

    UEdge uEdgeFromId(int id) const {
      return Parent::edgeFromId(id >> 1);
    }

    Node fromId(int id, Node) const {
      return nodeFromId(id);
    }

    Edge fromId(int id, Edge) const {
      return edgeFromId(id);
    }

    UEdge fromId(int id, UEdge) const {
      return uEdgeFromId(id);
    }


    Edge findEdge(Node source, Node target, Edge prev) const {
      if (prev == INVALID) {
	UEdge edge = Parent::findEdge(source, target);
	if (edge != INVALID) return direct(edge, true);
	edge = Parent::findEdge(target, source);
	if (edge != INVALID) return direct(edge, false);
      } else if (direction(prev)) {
	UEdge edge = Parent::findEdge(source, target, prev);
	if (edge != INVALID) return direct(edge, true);
	edge = Parent::findEdge(target, source);
	if (edge != INVALID) return direct(edge, false);	
      } else {
	UEdge edge = Parent::findEdge(target, source, prev);
	if (edge != INVALID) return direct(edge, false);	      
      }
      return INVALID;
    }

    UEdge findUEdge(Node source, Node target, UEdge prev) const {
      if (prev == INVALID) {
	UEdge edge = Parent::findEdge(source, target);
	if (edge != INVALID) return edge;
	edge = Parent::findEdge(target, source);
	if (edge != INVALID) return edge;
      } else if (Parent::source(prev) == source) {
	UEdge edge = Parent::findEdge(source, target, prev);
	if (edge != INVALID) return edge;
	edge = Parent::findEdge(target, source);
	if (edge != INVALID) return edge;	
      } else {
	UEdge edge = Parent::findEdge(target, source, prev);
	if (edge != INVALID) return edge;	      
      }
      return INVALID;
    }

  };


  template <typename _Base>
  class BpUGraphExtender : public _Base {
  public:
    typedef _Base Parent;
    typedef BpUGraphExtender Graph;

    typedef typename Parent::Node Node;
    typedef typename Parent::Edge UEdge;

    using Parent::first;
    using Parent::next;

    using Parent::id;

    Node source(const UEdge& edge) const {
      return aNode(edge);
    }
    Node target(const UEdge& edge) const {
      return bNode(edge);
    }

    void firstInc(UEdge& edge, bool& direction, const Node& node) const {
      if (Parent::aNode(node)) {
	Parent::firstOut(edge, node);
	direction = true;
      } else {
	Parent::firstIn(edge, node);
	direction = static_cast<UEdge&>(edge) == INVALID;
      }
    }
    void nextInc(UEdge& edge, bool& direction) const {
      if (direction) {
	Parent::nextOut(edge);
      } else {
	Parent::nextIn(edge);
	if (edge == INVALID) direction = true;
      }
    }

    int maxUEdgeId() const {
      return Parent::maxEdgeId();
    }

    UEdge uEdgeFromId(int id) const {
      return Parent::edgeFromId(id);
    }

    class Edge : public UEdge {
      friend class BpUGraphExtender;
    protected:
      bool forward;

      Edge(const UEdge& edge, bool _forward)
	: UEdge(edge), forward(_forward) {}

    public:
      Edge() {}
      Edge (Invalid) : UEdge(INVALID), forward(true) {}
      bool operator==(const Edge& i) const {
	return UEdge::operator==(i) && forward == i.forward;
      }
      bool operator!=(const Edge& i) const {
	return UEdge::operator!=(i) || forward != i.forward;
      }
      bool operator<(const Edge& i) const {
	return UEdge::operator<(i) || 
	  (!(i.forward<forward) && UEdge(*this)<UEdge(i));
      }
    };

    void first(Edge& edge) const {
      Parent::first(static_cast<UEdge&>(edge));
      edge.forward = true;
    }

    void next(Edge& edge) const {
      if (!edge.forward) {
	Parent::next(static_cast<UEdge&>(edge));
      }
      edge.forward = !edge.forward;
    }

    void firstOut(Edge& edge, const Node& node) const {
      if (Parent::aNode(node)) {
	Parent::firstOut(edge, node);
	edge.forward = true;
      } else {
	Parent::firstIn(edge, node);
	edge.forward = static_cast<UEdge&>(edge) == INVALID;
      }
    }
    void nextOut(Edge& edge) const {
      if (edge.forward) {
	Parent::nextOut(edge);
      } else {
	Parent::nextIn(edge);
        edge.forward = static_cast<UEdge&>(edge) == INVALID;
      }
    }

    void firstIn(Edge& edge, const Node& node) const {
      if (Parent::bNode(node)) {
	Parent::firstIn(edge, node);
	edge.forward = true;	
      } else {
	Parent::firstOut(edge, node);
	edge.forward = static_cast<UEdge&>(edge) == INVALID;
      }
    }
    void nextIn(Edge& edge) const {
      if (edge.forward) {
	Parent::nextIn(edge);
      } else {
	Parent::nextOut(edge);
	edge.forward = static_cast<UEdge&>(edge) == INVALID;
      }
    }

    Node source(const Edge& edge) const {
      return edge.forward ? Parent::aNode(edge) : Parent::bNode(edge);
    }
    Node target(const Edge& edge) const {
      return edge.forward ? Parent::bNode(edge) : Parent::aNode(edge);
    }

    bool direction(const Edge& edge) const {
      return edge.forward;
    }

    Edge direct(const UEdge& edge, const Node& node) const {
      return Edge(edge, node == Parent::source(edge));
    }

    Edge direct(const UEdge& edge, bool direction) const {
      return Edge(edge, direction);
    }

    Node oppositeNode(const UEdge& edge, const Node& node) const {
      return source(edge) == node ? 
	target(edge) : source(edge);
    }

    Edge oppositeEdge(const Edge& edge) const {
      return Edge(edge, !edge.forward);
    }

    int id(const Edge& edge) const {
      return (Parent::id(edge) << 1) + (edge.forward ? 0 : 1);
    }
    Edge edgeFromId(int id) const {
      return Edge(Parent::fromId(id >> 1, UEdge()), (id & 1) == 0);
    }
    int maxEdgeId() const {
      return (Parent::maxId(UEdge()) << 1) + 1;
    }

    class ANode : public Node {
      friend class BpUGraphExtender;
    public:
      ANode() {}
      ANode(const Node& node) : Node(node) {
	LEMON_ASSERT(Parent::aNode(node) || node == INVALID, 
		     typename Parent::NodeSetError());
      }
      ANode(Invalid) : Node(INVALID) {}
    };

    void first(ANode& node) const {
      Parent::firstANode(static_cast<Node&>(node));
    }
    void next(ANode& node) const {
      Parent::nextANode(static_cast<Node&>(node));
    }

    int id(const ANode& node) const {
      return Parent::aNodeId(node);
    }

    class BNode : public Node {
      friend class BpUGraphExtender;
    public:
      BNode() {}
      BNode(const Node& node) : Node(node) {
	LEMON_ASSERT(Parent::bNode(node) || node == INVALID,
		     typename Parent::NodeSetError());
      }
      BNode(Invalid) : Node(INVALID) {}
    };

    void first(BNode& node) const {
      Parent::firstBNode(static_cast<Node&>(node));
    }
    void next(BNode& node) const {
      Parent::nextBNode(static_cast<Node&>(node));
    }
  
    int id(const BNode& node) const {
      return Parent::aNodeId(node);
    }



    int maxId(Node) const {
      return Parent::maxNodeId();
    }
    int maxId(BNode) const {
      return Parent::maxBNodeId();
    }
    int maxId(ANode) const {
      return Parent::maxANodeId();
    }
    int maxId(Edge) const {
      return maxEdgeId();
    }
    int maxId(UEdge) const {
      return maxUEdgeId();
    }


    Node fromId(int id, Node) const {
      return Parent::nodeFromId(id);
    }
    ANode fromId(int id, ANode) const {
      return Parent::fromANodeId(id);
    }
    BNode fromId(int id, BNode) const {
      return Parent::fromBNodeId(id);
    }
    Edge fromId(int id, Edge) const {
      return edgeFromId(id);
    }
    UEdge fromId(int id, UEdge) const {
      return uEdgeFromId(id);
    }

  };

}

#endif // LEMON_UNDIR_GRAPH_EXTENDER_H
